#ifndef __RSTREAM_H
#define __RSTREAM_H

// DESCRIPTION: This file contains the implementation of a random
//              stream. The random stream is obtained by using the
//              linear congruential generator theory for which,
//              starting from a modulus m > 0, a multiplier a >= 0
//              and < m, an increment c >= 0 and and < m and an odd
//              seed X(0), a random stream is generated in the
//              following way:
//                             X(i+1)  =  X(i)*a +c(mod m).
//              In this case: m = 2**(32), a = 663608941 and c = 0.
//              The seed has to be given in input to the class
//              constructor, thus allowing creation of several 
//              different random streams.
//
//              NOTE: It is assumed that the arithmetic for the
//                    unsigned long is modulus 2**(32).
//
//              Sample points may be generated from a uniform 
//              distribution between (0,1); or from a uniform
//              distribution between user-specified lower and upper
//              bounds over discrete or continuous range; from a
//              Poisson distribution; from a negative exponential
//              distribution; or from a Pareto distribution.
//
//              Any of these distributions may be sampled from
//              independent streams.
//

#include <cmath>
#include <iostream>

using namespace std;

class Rstream
{
public:
  Rstream(unsigned long int = 0);
  void initStream(unsigned long int);
  const unsigned long int get_xi() const;
  double  uniform01();
  double uniform(double, double);
  double floor_uniform(double, double);
  double Exponential(double);
  long UniformLong(long, long);
  double unit_rate_poisson(float);
  double Pareto(double,double);
  short Trimodal_Internet();
  double RandomGaussian(double,double);

private:
  // Value of the last random number generated (seed)
  unsigned long       xi; 

  float                   lngamma(float) const;
};

#endif


