/*
 simengine.h and simengine.cpp implement a simple yet easy to use and powerful
 discrete event simulation engine. The code was written by Gergely Zaruba
 (zaruba@ieee.org) based on an original by Andrew Myers.

 This code is provided to you under general GNU licence assumptions.
 Please enjoy it!
*/
/*

- HANDLING EVENTS
  Each event type has to be handled with a funztion inside your object ,e.g., 
  if your object is called My_Object, then the function could handle the event type:

  void My_Object::Event_type_A_handler();

  (note, that you can change the name of this function to whatever you wish it to)
  if Event_type_A_handler() is indeed the function to be handling event type A, then there 
  needs to be (mandatory) an additional static version of this function, starting with with a 
  a void * parameter (following with parameters that Event_type_A_handler had (see argument list
  below). The purpose of this static function is to provide a hook to the non-staic 
  function, thus it will need to call the non-static function, e.g.,
  
void My_Object::S_Event_type_A_handler(void * c_a)
  {
    ((My_Object*)c_a)->Event_type_A_handler();
  }


  This is a tricky way to fooling C++ and being able to call non-static member variables of
  objects from the outside. Note, that both these functions will have to be prototyped in
  your header files (as usual).


  Example-2:
  
    void My_Object::S_Event_type_A_handler(void * c_a, long c_a, long c_b)
    {
      ((My_Object*)c_a)->Event_type_A_handler(c_a,c_b);
    }

    void My_Object::Event_type_A_handler(long c_a, long c_b)
      {
        //here comes the code to execute when a scheduled type A event hits
	// you can use c_a, and c_b to pass over variables through the simulation engine
	// to this handler function
      }

- SCHEDULING EVENTS
  Events can be scheduled with relative time "offsets" from the current simulation time. 
  To schedule an event of type A in exactly duration second from now on (assuming that
  this type of event is handled by Event_type_A_handler() the following command may be used:

  new DSEvent(this, duration, S_Event_type_A_handler);

  Example-4:
    Taking the functions from Example-2 an event may be scheduled:
     
      new DSEvent(this, duration, S_Event_type_A_handler, long_i_want_to_pass, long2_i_want_to_pass);

- ARGUMENTS OF EVENT HANDLERS:
  Event handler functions can have one from the following set of arguments: 
    - void
    - long 
    - long, long
    - double
    - double, double
    - long, double
    - long, long, double
    - long, double, double
    - long, long, double, double
    (if other arguments need to be supported the simengine.h and simengine.cpp
    files will have to be modified accordingly) 


- TO START THE SIMULATION:
          Simengine().Simulate(simulationtime); 

- TO OBTAIN CURRENT SIMULATION TIME:
          Simengine().SimTime();

*/

#ifndef _SIM_H_
#define _SIM_H_

#include <iostream>
#include <set>


using namespace std;

struct SimEvent {
public:
  SimEvent(double);

  virtual void Destroy();
  virtual void Execute();
  double EventTime() const;
  inline SimEvent* MyPointer(){return this;};
protected:
  SimEvent(const SimEvent &);
  SimEvent & operator=(const SimEvent &);
  virtual ~SimEvent();

  double timeOfOccurrence;
};

struct EndSimulation : public SimEvent {
public:
  EndSimulation(double = 0);

  virtual void Destroy();
  virtual void Execute();
protected:
  EndSimulation(const EndSimulation &);
  EndSimulation & operator=(const EndSimulation &);
  virtual ~EndSimulation();
};

struct CompareEvents {
public:
  bool operator()(const SimEvent *e1, const SimEvent *e2) const {
    return e1->EventTime() < e2->EventTime();
  }
};

struct DiscreteEventSim {
public:
  void Simulate(double);        // Run simulator
  void InsertEvent(SimEvent *); // Add event to be executed
  void DeleteEvent(SimEvent *); // Prevent event from execution

  double SimTime() const; // Get current simulation time
private:
  DiscreteEventSim();
  DiscreteEventSim(const DiscreteEventSim &);
  DiscreteEventSim & operator=(const DiscreteEventSim &);

  void TerminateSimulation();

  multiset<SimEvent *, CompareEvents> eventQ;
  long eventsExecuted;
  double currentTime;
  double finishTime;
  bool terminate;

  friend struct DiscreteEventSim & Simengine();
  friend void EndSimulation::Execute();
};

struct DSEvent : public SimEvent {

    DSEvent(void*, double, void(void *));  //type:0
    DSEvent(void*, double, void(void *, long), long); //type:1
    DSEvent(void*, double, void(void *, long, long), long, long); //type:2
    DSEvent(void*, double, void(void *, double), double); //type:3
    DSEvent(void*, double, void(void *, double, double), double, double); //type:4
    DSEvent(void*, double, void(void *, long, double), long, double); //type:5
    DSEvent(void*, double, void(void *, long, long, double), long, long, double); //type:6
    DSEvent(void*, double, void(void *, long, double, double), long, double, double); //type:7
    DSEvent(void*, double, void(void *, long, long, double, double), long, long, double, double); //type:8
    virtual void Destroy();
    virtual void Execute();
  protected:
    DSEvent(const DSEvent &);
    DSEvent & operator=(const DSEvent &);
    virtual ~DSEvent();

    void (*myfunct0)(void*);
    void (*myfunct1)(void*, long);
    void (*myfunct2)(void*, long, long);
    void (*myfunct3)(void*, double);
    void (*myfunct4)(void*, double, double);
    void (*myfunct5)(void*, long, double);
    void (*myfunct6)(void*, long, long, double);
    void (*myfunct7)(void*, long, double, double);
    void (*myfunct8)(void*, long, long, double, double);

    long mylparam1;
    long mylparam2;
    double mydparam1;
    double mydparam2;
    char functype;
    void * s_o_ptr;
  };


struct DiscreteEventSim & Simengine();

#endif
